/*!*****************************************************
* @file
* @brief FitSqr is a class for quadratic polynomial and Pade type fits to nonlinear curves.
*
* For the mathematical background see
* <a href="http://www.mpia-hd.mpg.de/~mathar/mathar20121207.pdf">Fit to initially linear section with saturation</a>
*/

/*
* $Header: https://svn.mpia.de/gulli/geirs/src/trunk/extern/fits2fwhm/FitSqr.cxx 798 2021-06-15 13:53:57Z mathar $
*/

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <iostream>

/* temporarily for sleep(3)
#include <unistd.h>
*/

#include "FitSqr.h"

#ifdef HAVE_GSL_GSL_LINALG_H
#include <gsl/gsl_sf.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_linalg.h>
#endif

/** Determinant of the 2 by 2 matrix */
#define DET22(a00,a01,a10,a11) ((a00)*(a11)-(a01)*(a10))

/** Determinant of a 3by3 matrix
* @param[in] matr The elements of the matrix
* @return The determinant of the matrix.
* @author Richard J. Mathar
*/
double FitSqr::det33(const double matr[3][3])
{
	/* expand along column 0 with the checkerboard rule
	*/
	return matr[0][0]*DET22(matr[1][1],matr[1][2],matr[2][1],matr[2][2])
		- matr[1][0]*DET22(matr[0][1],matr[0][2],matr[2][1],matr[2][2]) 
		+ matr[2][0]*DET22(matr[0][1],matr[0][2],matr[1][1],matr[1][2]) ;
} /* det33 */

/** Solve a 3 by 3 system of linear equations.
* @param[in] A The 3 by 3 coefficient matrix.
* @param[in] rhs The vector of the right hand side (inhomogeneous term)
* @param[out] res The resulting vector
* @since 2012-12-06
* @author Richard J. Mathar
*/
void FitSqr::solve33(const double A[3][3], const double rhs[3], double res[3])
{
	/* denominator in Cramer's rule
	*/
	const double cramdenom = det33(A) ;

	/* Matrix with the substituted columns for Cramer's rule
	*/
	double cram[3][3] ;

	/* determine the three solutions alpha by Cramer's rule
	*/
	for(int c=0 ; c < 3 ; c++)
	{
		memcpy(cram,A,9*sizeof(double)) ;
		/* push right hand side into column c of the matrix
		*/
		for(int r=0 ; r< 3; r++)
			cram[r][c] = rhs[r] ;
		res[c] = det33(cram)/ cramdenom ;
	}
} /* solve33 */

/** Solve a 2 by 2 system of linear equations.
* @param[in] A The 2 by 2 coefficient matrix.
* @param[in] rhs The vector of the right hand side (inhomogeneous term)
* @param[out] res The resulting vector
* @since 2015-05-26
* @author Richard J. Mathar
*/
void FitSqr::solve22(const double A[2][2], const double rhs[2], double res[2])
{
	/* denominator in Cramer's rule
	*/
	const double cramdenom = DET22(A[0][0],A[0][1],A[1][0],A[1][1]) ;

	/* Matrix with the substituted columns for Cramer's rule
	*/
	double cram[2][2] ;

	/* determine the two solutions alpha by Cramer's rule
	*/
	for(int c=0 ; c < 2 ; c++)
	{
		memcpy(cram,A,4*sizeof(double)) ;
		/* push right hand side into column c of the matrix
		*/
		for(int r=0 ; r< 2; r++)
			cram[r][c] = rhs[r] ;
		res[c] = DET22(cram[0][0],cram[0][1],cram[1][0],cram[1][1])/ cramdenom ;
	}
} /* solve22 */

/** Invert a 3 by 3 matrix.
* This is a slow implementation which solves 3 inhomogeneous linear systems in succession.
* Singularity is not checked in any form.
* @param[in] A The 3 by 3 coefficient matrix to be inverted.
* @param[out] inv The resulting inverse matrix.
* @since 2012-12-06
* @author Richard J. Mathar
*/
void FitSqr::invert33(const double A[3][3], double inv[3][3])
{
	/* denominator in Cramer's rule
	*/
	double cramdenom = det33(A) ;

	/* Matrix with the substituted columns for Cramer's rule
	*/
	double cram[3][3] ;

	for(int p=0 ;p < 3 ; p++)
	{
		/* determine the three solutions alpha by Cramer's rule
		*/
		for(int c=0 ; c < 3 ; c++)
		{
			memcpy(cram,A,9*sizeof(double)) ;
			/* push right hand side into column c of the matrix
			*/
			for(int r=0 ; r< 3; r++)
			{
				cram[r][c] = (r==p) ? 1.0 : 0.0 ;
			}
			inv[c][p] = det33(cram)/ cramdenom ;
		}
	}
} /* invert33 */

/** Compute the sum of the squared differences relative to the fit. 
* @param[in] x The vector of the abscissae of the measurements
* @param[in] y The vector of the noisy ordinates of the measurements.
* @param[in] N The number of entries in x and also in y.
* @param[in] alpha The coefficients of the linear quadratic fit.
* @param[in] p The total power of each component of the x-vector, >=0.
* @return The sum_i x[i]^p*(difference y[i] from the fitted y[i])^2.
* @author Richard J. Mathar
*/
double FitSqr::sqrVar(const double x[], const double y[], int N, double alpha[3], int p)
{
	double v=0. ;
	switch(p)
	{
	case 0 :
		for (int i=0 ; i < N ; i++)
			v += pow( alpha[0]+x[i]*(alpha[1]+alpha[2]*x[i])-y[i],2.) ;
		break;
	default :
		for (int i=0 ; i < N ; i++)
			v += pow(x[i],(double)p)*pow(alpha[0]+x[i]*(alpha[1]+alpha[2]*x[i])-y[i],2.) ;
		break;
	}
	return v;
} /* sqrVar */

/** Fit a set of scattered data points (x,y) with a quadratic polynomial.
* @param[in] x The vector of all x-values. has N compontents.
* @param[in] y The vector of all y-values. has N compontents.
* @param[in] N The number of components individually in x and y.
* @param[out] alpha The three coefficients of the quadratic fit.
* @param[out] alphaVar The main diagonal of the variance matrix of the alpha.
*  Not computed (for speed) if NULL.
* @return Sum of the squares of the deviations from the fit.
* @author Richard J. Mathar
*/
double FitSqr::fitsqr(const double x[], const double y[], int N, double alpha[3], double alphaVar[3])
{
	/* 3 by 3 matrix of the inhomogeneous linear system of eqs
	*/
	double A[3][3] ;

	/* right hand side of the system of equations
	*/
	double rhs[3] ;

	A[0][0] = N ;
	A[0][1] = A[0][2] = A[1][2] = A[2][2]=0. ;
	rhs[0] = rhs[1] = rhs[2] =0. ;
	/* sum over all x_i and y_i */
	for (int i=0 ; i < N ; i++)
	{
		const double xsq = x[i]*x[i] ;
		A[0][1] += x[i] ;
		A[0][2] += xsq ;
		A[1][2] += x[i]*xsq ;
		A[2][2] += xsq*xsq ;
		rhs[0] += y[i] ;
		rhs[1] += y[i]*x[i] ;
		rhs[2] += y[i]*xsq ;
	}
	A[1][0] = A[0][1] ;
	A[1][1] = A[2][0] = A[0][2] ;
	A[2][1] = A[1][2] ;

	/* solve the inhomogeneous linear system
	*/
	solve33(A,rhs,alpha) ;
#if 0
	/* typically happens if there are less than 2 values in the data set...
	*/
	if (isnan(alpha[0]) || isnan(alpha[1]) || isnan(alpha[2]) )
		std::cerr <<__FILE__ << " " << __LINE__ << " " << A[0][0] << " " 
		<< A[0][1] << " " << A[0][2] << " " << A[1][1] << " " << A[1][2] << std::endl ;
#endif
	if ( alphaVar )
	{
		/* Inverse of the 3 by 3 matrix of the inhomogeneous linear system of eqs
		*/
		double Ainv[3][3] ;

		/* variances of the right hand side of the system of equations
		*/
		double rhsVar[3][3] ;

		/* Calculate the variances by multiplication of the variances of the
		* rhs by the inverse matrix. Uses the symmetry in the covariance matrix.
		*/
		invert33(A,Ainv) ;
	
		for(int r=0 ; r < 3; r++)
		for(int c=r ; c < 3; c++)
			rhsVar[c][r] = rhsVar[r][c] = sqrVar(x, y, N, alpha, r+c) ;
#if 0
		/* print the covariance matrix. Not useful in general
		*/
		printf("rhsVar %e %e %e\n",rhsVar[0][0],rhsVar[0][1],rhsVar[0][2]) ;
		printf("rhsVar %e %e %e\n",rhsVar[1][0],rhsVar[1][1],rhsVar[1][2]) ;
		printf("rhsVar %e %e %e\n",rhsVar[2][0],rhsVar[2][1],rhsVar[2][2]) ;
		/* print the inverse matrix. Not useful in general
		*/
		printf("inv %e %e %e\n",Ainv[0][0],Ainv[0][1],Ainv[0][2]) ;
		printf("inv %e %e %e\n",Ainv[1][0],Ainv[1][1],Ainv[1][2]) ;
		printf("inv %e %e %e\n",Ainv[2][0],Ainv[2][1],Ainv[2][2]) ;
#endif
		/* take only care of the main diagonal of the covariance. No PVD yet.
		*/
		for(int r=0 ; r < 3; r++)
		{
			alphaVar[r] = 0. ;
			for(int i=0 ; i < 3; i++)
			for(int j=0 ; j < 3; j++)
				alphaVar[r] += Ainv[r][i]*Ainv[r][j]*rhsVar[i][j] ;
		}
	
		/* return the sum of all squared deviations
		*/
		return rhsVar[0][0] ;
	}
	else
		return 0. ;
} /* fitsqr */

/** 
* @param[in] x The vector of all x-values. has N compontents.
* @param[in] y The vector of all y-values. has N compontents.
* @param[in] N The number of components individually in x and y.
* @param[out] alpha The two coefficients of the quadratic fit.
* @return Always zero.
* @author Richard J. Mathar
* @since 2015-05-26
* @since 2015-12-02 deprecated and replaced by the vesion with the aducut parameter
*/
double FitSqr::fitlin(const double x[], const double y[], int N, double alpha[2])
{
#ifdef HAVE_GSL_GSL_LINALG_H
	/* linear fit: means 2 coefficients 
	* It might be much faster to use the more explicit code further down, even
	* if the gsl library is available.
	*/
	return fitpoly(x,y,N,2,alpha) ;
#else
	/* 2 by 2 matrix of the inhomogeneous linear system of eqs
	*/
	double A[2][2] ;

	/* right hand side of the system of equations
	*/
	double rhs[2] ;

	A[0][0] = N ;
	A[0][1] = A[1][1] = 0. ;
	rhs[0] = rhs[1] =0. ;
	/* sum over all x_i and y_i */
	for (int i=0 ; i < N ; i++)
	{
		A[0][1] += x[i] ;
		A[1][1] += x[i]*x[i] ;
		rhs[0] += y[i] ;
		rhs[1] += y[i]*x[i] ;
	}
	A[1][0] = A[0][1] ;

	/* solve the inhomogeneous linear system
	*/
	solve22(A,rhs,alpha) ;

	return 0. ;
#endif /* HAVE_GSL_GSL_LINALG_H */
} /* fitlin */

/** 
* @param[in] x The vector of all x-values. has N compontents.
* @param[in] y The vector of all y-values. has N compontents.
* @param[in] N The number of components individually in x and y.
* @param[out] alpha The two coefficients of the quadratic fit.
* @return Always zero.
* @author Richard J. Mathar
* @since 2015-05-26
*/
double FitSqr::fitlin(const double x[], const double y[], int N, double alpha[2], const int aducut)
{
	/* 2 by 2 matrix of the inhomogeneous linear system of eqs
	*/
	double A[2][2] ;

	/* right hand side of the system of equations
	*/
	double rhs[2] ;

	A[0][1] = A[1][1] = 0. ;
	rhs[0] = rhs[1] =0. ;
	/* sum over all x_i and y_i, but include only the values below the aducut */
	int i=0 ;
	for ( ; i < N ; i++)
	{
		/* skip saturated values
		*/
		if ( y[i] >= aducut)
			break;
		A[0][1] += x[i] ;
		A[1][1] += x[i]*x[i] ;
		rhs[0] += y[i] ;
		rhs[1] += y[i]*x[i] ;
	}
	A[1][0] = A[0][1] ;
	A[0][0] = i ;

	if ( i > 1)
		/* solve the inhomogeneous linear system if there is a slope to be drived from at least 2 values
		*/
		solve22(A,rhs,alpha) ;
	else
		/* otherwise decline to fit and set pixel to zero
		*/
		alpha[0]=alpha[1]=0.0 ;

	return 0. ;
} /* fitlin */

#ifdef HAVE_GSL_GSL_LINALG_H
/** 
* @param[in] x The vector of all x-values. has N compontents.
* @param[in] y The vector of all y-values. has N compontents.
* @param[in] N The number of components individually in x and y.
* @param[in] deg The number of unknown coefficients
* @param[out] alpha The unknown coefficients of the fit.
* @return Always zero.
* @author Richard J. Mathar
* @since 2015-05-26
*/
double FitSqr::fitpoly(const double x[], const double y[], int N, int deg, double alpha[])
{
	/* the deg by deg matrix of the linear system of equations
	* apparently the calloc() functions lead to a much higher probability
	* of running into singular matrices below (??)
	*/
#if 1
	gsl_matrix *A=gsl_matrix_alloc(deg,deg) ;
	gsl_matrix_set_zero(A) ;
#else
	gsl_matrix *A=gsl_matrix_calloc(deg,deg) ;
#endif

	/* the RHS of the linear system of equations */
#if 1
	gsl_vector *rhs=gsl_vector_alloc(deg) ;
	gsl_vector_set_zero(rhs) ;
#else
	gsl_vector *rhs=gsl_vector_calloc(deg) ;
#endif

	/* fill the upper right portion of the matrix and the right hand side
	*/
	for(int row=0; row<deg ;row++)
	{
		for(int col= row; col< deg ;col++)
		{
			for(int k=0 ; k < N ;k++)
				* gsl_matrix_ptr(A,row,col) += pow(x[k],row+col) ;
		}
		for(int k=0 ; k < N ;k++)
			*gsl_vector_ptr(rhs,row) += pow(x[k],row)*y[k] ;
        }

	/* apply symmetry for the lower triangular submatrix */
	for(int row=1; row<deg ;row++)
	for(int col=0; col<row ;col++)
		* gsl_matrix_ptr(A,row,col) = * gsl_matrix_ptr(A,col,row) ;

	/* avoid core dumps by sparse numbers of matrices with zero determinant
	*/
	gsl_set_error_handler_off() ;

	/* solve the linear system of equations
	* return value of 21 means singularity see gsl-2.1/err/gsl_errno.h
	*/
	const int gslstat = gsl_linalg_HH_svx(A,rhs) ;

	if ( gslstat )
	{
		for(int row=0; row < deg ; row++)
			alpha[row] = 0.0 ;
	}
	else
	{
		/* pack the expansion coefficients back into alpha
		*/
		for(int row=0; row < deg ; row++)
			alpha[row] = gsl_vector_get(rhs,row) ;
	}

	gsl_vector_free(rhs) ;
	gsl_matrix_free(A) ;

	return 0 ;

} /* fitpoly */
#endif /* HAVE_GSL_GSL_LINALG_H */

/** Fit a set of scattered data points (x,y) to a first-order rational polynomial.
* @param[in] x The vector of all x-values. has N compontents.
* @param[in] y The vector of all y-values. has N compontents.
* @param[in] N The number of components individually in x and y.
* @param[out] alpha The three coefficients of the fit.
*  The points (xi,yi) are fitted by y= (alpha[0]+alpha[1]*x)/(1+alpha[2]*x).
* @return Sum of the squares of the deviations from the fit.
* @author Richard J. Mathar
*/
double FitSqr::fitsat(const double x[], const double y[], int N, double alpha[3])
{
	/* 3 by 3 matrix of the inhomogeneous linear syste of eqs
	*/
	double A[3][3] ;

	/* right hand side of the system of equations
	*/
	double rhs[3] ;

	/* denominator in Cramer's rule
	*/
	double cramdenom ;

	/* Matrix with the substituted columns for Cramer's rule
	*/
	double cram[3][3] ;

	/* relaxation factor in the newton-ralphson iterations
	*/
	double rlxf = 0.30 ;

	/* re-interpret the coefficients of the quadratic fit
	* in terms of the rational function fit
	*/
	alpha[2]= -alpha[2]/alpha[1] ;
	alpha[1] += alpha[0]*alpha[2] ;

#ifdef DEBUG
		printf("# start %e %e %e\n",alpha[0],alpha[1],alpha[2]) ;
#endif

	for(int itr =0 ; itr<30; itr++)
	{
		for(int r=0 ; r < 3 ; r++)
			A[r][0] = A[r][1] = A[r][2] = rhs[r] = 0. ;


		/* to stabilize the method, use only half, then 3/4 and eventually
		* all of the data values to fit the three parameters.
		*/
		int Neff = ( itr < 4 ) ?  N/2 : ( itr< 8 ? 3*N/4 : N) ;
		/* build matrix and right hand side by summation over all noisy values
		*/
		for (int i=0 ; i < Neff ; i++)
		{
			const double xinv = 1/(1+alpha[2]*x[i]) ;
			const double xinv2 = xinv*xinv ;
			const double xlin = alpha[0]+alpha[1]*x[i] ;
			const double xlinr = xlin*xinv-y[i] ;
			rhs[0] -= xlinr*xinv ;
			rhs[1] -= xlinr*xinv*x[i] ;
			rhs[2] += xlinr*xlin*x[i]*xinv2 ;
			A[0][0] += xinv2 ;
			A[0][1] += x[i]*xinv2 ;
			A[0][2] += x[i]*xinv2*(-2.*xlin*xinv+y[i]) ;
			A[1][1] += x[i]*x[i]*xinv2 ;
			A[1][2] += x[i]*x[i]*xinv2*(-2.*xlin*xinv+y[i]) ;
			A[2][2] += x[i]*x[i]*xlin*xinv2*(3.*xlin*xinv2-2*y[i]*xinv) ;
		}
		A[1][0] = A[0][1] ;
		A[2][0] = A[0][2] ;
		A[2][1] = A[1][2] ;
#ifdef DEBUG
		printf("# zero ? %e %e %e\n",rhs[0],rhs[1],rhs[2]) ;
#endif
	
		/* denominator in Cramer's rule
		*/
		cramdenom = det33(A) ;

		/* determine the three solutions alpha by Cramer's rule
		*/
		for(int c=0 ; c < 3 ; c++)
		{
			memcpy(cram,A,9*sizeof(double)) ;
			/* push right hand side into column c of the matrix
			*/
			for(int r=0 ; r< 3; r++)
				cram[r][c] = rhs[r] ;
			alpha[c] += rlxf *det33(cram)/ cramdenom ;
		}
		if ( itr > 15)
			rlxf = 0.5 ;

#ifdef DEBUG
		/* if in debugging mode, print the some progress
		*/
		printf("# alph = %e %e %e rl= %f\n",alpha[0],alpha[1], alpha[2],rlxf) ;
#endif
	}
	
	/* calculate the sum of all squared deviations
	*/
	cramdenom=0. ;
	for(int i=0 ; i < N ; i++)
	{
		cramdenom += pow( (alpha[0]+x[i]*alpha[1])/(1.+alpha[2]*x[i])-y[i],2.) ;
#ifdef DEBUG
		/* if in debugging mode, print the original values and their
		* associated values on the fitting curve, suitable for gnuplot.
		*/
		printf("%e %e %e\n",x[i],y[i], (alpha[0]+x[i]*alpha[1])/(1.+alpha[2]*x[i]) ) ;
#endif
	}
	return cramdenom ;
} /* fitsat */
