/*!***************************************************
* @file
* @brief Interface to the extraction of seeing parameters via the sextractor of astromatic.net.
*
* $Header: https://svn.mpia.de/gulli/geirs/src/trunk/extern/fits2fwhm/Fits2fwhm.cxx 800 2022-05-13 14:46:24Z mathar $
*/

#include <cstdio>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstring>
#include <limits>
#include <cmath>
#include <algorithm>
#include <errno.h>
#include <unistd.h>
#include <fitsio.h>

#include "Fits2fwhm.h"

using namespace std ;

/**
* @brief the FITS header keyword which contains the binning factor
*/
#define FITS2FWHM_BINKW "XBINNING"

/**
* @brief the FITS header keyword which contains the pixel scale (unbinned)
*/
#define FITS2FWHM_PIXSCKW "PIXSCALE"

/*!**************************************** 
* @brief ctor
* @param[in] chop The number of pixels (4 sided frames) to be ignored (chopped) in the original images.
* @param[in] fockw The HDU keyword which gives the focal plane position.
*     In the fits files the fockw =... values provide the abscissae of the least squares algorithm.
* @param[in] bat A 'batch job' flag to indicate non-interactive use.
*     This essentially means no GUI's pop up and everything produces data in stdout.
* @since 2020-03
* @author Richard J. Mathar
*/
Fits2fwhm::Fits2fwhm(int chop, const string & fockw, bool bat, float thold)  
	: choppx(chop), focpos(fockw), batch(bat), thresh(thold)
{
} /* Fits2fwhm::Fits2fwhm */

/*!**************************************** 
* @brief The function which will be called intermediately through Application::start()
* @param argc The number of strings in argv[]
* @param argv The FITS file names of the command line.
* @return An integer as a status value. 0 if successful.
* @since 2020-03
* @author Richard J. Mathar
*/
int Fits2fwhm::run(int argc, char *argv[])
{
	int status=0 ;

	/* call sex(1) and extract the centroids. The result
	* is a list of FITS pixel coordinates in the pcam image.
	*/
	
	if ( argc == 1 )
	{
		/* this is a single FITS file which we examine in full,
		* extracting all the widths and displaying a ds9 image
		* Temporary file name for the FITS file
		*/
		char *skfits = new char[strlen(__FILE__)+12] ;
		sprintf(skfits,"%sXXXXXX.fits",__FILE__) ;
		::mkstemps(skfits,5) ;
	
		/* remove empty file because saveFits needs an exclamation mark to overwrite
		*/
		remove(skfits) ;

		/* create temporary chopped image file. Try to get the XBINNING value.
		*/
		int bin = 0 ;
		float pixscal ;
		float dummy ;
		int err = chopFrame(string(argv[0]), skfits, bin,pixscal,dummy) ;

		if ( err == 0 )
			/* run sex(1) on that chopped image
			*/
			centroids(skfits,bin,pixscal) ;
		else
			cerr << "Unable to create " << skfits << endl ;

		/* delete temporary fits file and associated name
		*/
		remove(skfits) ;
		delete[] skfits ;
	}
	else if ( argc >=3 )
	{
		/* these are three or more FITS files from which we extract
		* the T_FOCUS keyword and run a least squares fit to get
		* a best focal position.
		*/

		/* 
		* Temporary file name for the parameter file for the parabolic fit
		*/
		char *ffil = new char[strlen(__FILE__)+11] ;
		sprintf(ffil,"%sXXXXXX.dat",__FILE__) ;
		::mkstemps(ffil,5) ;

		std::ofstream off(ffil) ;
		/* the common binning factor i the images...
		*/
		int binAll = 0 ;
		/* loop over input files
		*/
		for(int f=0 ; f < argc ; f++)
		{
			/* this is a single FITS file which we examine in full,
			* extracting all the widths and displaying a ds9 image
			* Temporary file name for the FITS file
			*/
			char *skfits = new char[strlen(__FILE__)+12] ;
			sprintf(skfits,"%sXXXXXX.fits",__FILE__) ;
			::mkstemps(skfits,5) ;
		
			/* remove empty file because saveFits needs an exclamation mark to overwrite
			*/
			remove(skfits) ;
	
			/* create temporary chopped image file. Try to get the XBINNING value.
			*/
			int bin = 0 ;
			float pixscal ;
			float focval ;
			int err = chopFrame(string(argv[f]), skfits, bin,pixscal,focval) ;
	
			if ( err == 0 )
			{
				if ( isnan(focval) )
					std::cerr << "No " << focpos << " keyword in " << argv[f] << " .. skipped" << std::endl ;
				else
				{
					if ( f > 0 && bin != binAll)
						std::cerr << "WARNING Incompatible binning factor " << bin << " in " << argv[f] << std::endl ;
					binAll = bin ;
					/* run sex(1) on that chopped image
					*/
					vector<float> widths = centroids(skfits,bin,pixscal) ;

					/* Add the widths to the ASCII file.
					* Use a simple reduction of the number of data by sorting the widths[]
					* and removing the values more than 1sigma away. For a list of quantiles
					* see the perc[] array in ccdread3.cxx addPerc(). 1sigma=68.2percent inside,
					* 0.3173 eliminated/outside, 0.158 eliminated at low and high end...
					*/
					std::sort(widths.begin(), widths.end()) ;

					const int skipw = (int) (widths.size()*0.158) ;
					for(int s= skipw ; s < (int)widths.size()-skipw ; s++)
						off << focval << " " << widths[s] << endl ;
				}
			}
			else
				cerr << "Unable to create " << skfits << endl ;
	
			/* delete temporary fits file and associated name
			*/
			remove(skfits) ;
			delete[] skfits ;
		}
		off.close() ;

		/* run fitSqr on the file without the -q option, default columns
		*/
		ostringstream cmd ;
		cmd << "fitSqr " << ffil ;
		system(cmd.str().c_str()) ;

		/* delete temporary fits file and associated name
		*/
		remove(ffil) ;
		delete[] ffil ;
	}

	return status ;
	
} /* Fits2fwhm::run */

/*!***************************************
* @brief Remove a frame of pixels at the 4 borders of a FITS image.
*   The image of width X height in file is reduced to an image
*   of (width-2*chop) X (height -2 *chop) in the output file.
*   The output file is overwritten if it exists.
*   If the prameter chop is too large compared to width or height
*   of the original image, it is reduced by a factor 2 or 4 such
*   that there are pixels left in the output file. 
* @param[in] ifile the original FITS file
* @param[in] ofile the FITS file to be created with the reduced interior part.
* @param[out] bin Binning parameter taken from the XBINNING keyword
* @param[out] pxscal Unbinned pixel scale from the PIXSCALE keyword
* @param[out] focparm The value of the focus parameter taken from the FITS header
* @return 0 if successful, errno on error
* @todo: use fits_open_data instead fits_open_file to get naxes[] if there is no image in PHDU
*/
int Fits2fwhm::chopFrame(const string & ifile, const string & ofile, int & bin, float & pixscal, float & focparm)
{
	/* step1 : read NAXIS1 and NAXIS2, check that 2*chop is less than both,
	* compute [1+chop:NAXIS1-chop,1+chop:NAXIS2-chop]
	*/

	/* cfitsio status
	*/
	int status=0 ;

	/* cfitsio handler
	*/
	fitsfile *fptr =0;

        fits_open_file(& fptr,ifile.c_str(), READONLY, &status) ;

	if ( status)
	{
		/* end up here if the FITS file does not exist or is malformed...
		*/
		fits_report_error(stderr,status) ;
		return ENOENT ;
	}

	long naxes[2] ;
	fits_get_img_size(fptr,sizeof(naxes)/sizeof(long), naxes, &status) ;
	if ( status)
	{
		/* end up here if the FITS file does not exist or is very malformed...
		*/
		fits_report_error(stderr,status) ;
	        fits_close_file(fptr,&status) ;
		return EDOM ;
	}

	// cerr << naxes[0] << " " << naxes[1] << endl ; // debugging

	/* try to get the bin factor and pix scale; don't care if this fails...
	*/
	fits_read_key(fptr,TINT,FITS2FWHM_BINKW, &bin, NULL, &status) ;
	status =0 ;
	pxscal = PIXSCALE ; /* default from Makefile.am */
	fits_read_key(fptr,TFLOAT,FITS2FWHM_PIXSCKW, &pixscal, NULL, &status) ;
	status =0 ;
	// cerr <<__FILE__ << " " << __LINE__ << " " << bin << " " << pxscal << endl ; // debugging
	focparm = numeric_limits<float>::quiet_NaN() ;
	fits_read_key(fptr,TFLOAT,focpos.c_str(), &focparm, NULL, &status) ;
	status =0 ;

        fits_close_file(fptr,&status) ;
        if ( status)
                fits_report_error(stderr,status) ;

	if ( 2*choppx > naxes[0] || 2*choppx > naxes[1])
		choppx /= 2 ;
	if ( 2*choppx > naxes[0] || 2*choppx > naxes[1])
		choppx /= 2 ;

	/* step 2: run ftcopy infile[] ofile clobber=yes, chopping the input file
	*/
	ostringstream sel ;
#if 0
	sel << "ftcopy " << ifile ;
	if ( choppx > 0 )
	{
		sel << "'[" << (1+choppx) << ":" << (naxes[0]-choppx) 
			<< "," << (1+choppx) << ":" << (naxes[1]-choppx) << "]'" ;
	}
	/* variant which needs the ftcopy executable of heasarc in the PATH
	*/
	sel << " " << ofile ;
	sel << " copyall=NO" ;
	system(sel.str().c_str()) ;
#else
	/* variant which does the copy in the member function
	*/
	sel << ifile ;
	if ( choppx > 0 )
	{
		sel << "[" << (1+choppx) << ":" << (naxes[0]-choppx) 
			<< "," << (1+choppx) << ":" << (naxes[1]-choppx) << "]" ;
	}
	ftcopy(sel.str().c_str(),ofile.c_str(),0) ;
#endif
	return 0 ;
} /* Fits2fwhm::chopFrame */

/*!**************************************** 
* @brief Extract widths from the image via sex(1).
* @param[in] skifts The name of the FITS file to be examined.
* @param[in] bin The binning factor. Value less than 1 means unknown.
* @param[in] pxscal pixel scale (unbined)  in arcsec/pixel
* @param[in] thresh a detection threshold (relative).
*    A value of the order of 2. A larger value reduces the number of detected/extracted stars.
* @return a magnitude-ordered list of stellar widths
* @since 2020-03
* @todo exit early if sex(1) is not in the path.
* @author Richard J. Mathar
*/
vector<float> Fits2fwhm::centroids(const char *skfits, int bin, float pixscal)
{

	/* temporary file for the convolution mask; 6 pixels FWHM
	*/
	char *skconv = new char[strlen(__FILE__)+12] ;
	sprintf(skconv,"%sXXXXXX.conv",__FILE__) ;
	::mkstemps(skconv,5) ;

	std::ofstream of1(skconv) ;

#if 1
	/*
	printf("CONV NORM\n") :
	fwhm := 6.0 :
	for pxx from -5 to 5 do
	for pxy from -5 to 5 do
	        printf("%f ",exp(-4.0*log(2.0)*(pxx^2+pxy^2)/fwhm^2)) ;
	end do:
	printf("\n") ;
	end do:
	*/
	of1 << "CONV NORM" << std::endl ;
	of1 << "0.021262 0.042525 0.072908 0.107155 0.135007 0.145816 0.135007 0.107155 0.072908 0.042525 0.021262" << std::endl ;
	of1 << "0.042525 0.085049 0.145816 0.214311 0.270015 0.291632 0.270015 0.214311 0.145816 0.085049 0.042525" << std::endl ;
	of1 << "0.072908 0.145816 0.250000 0.367434 0.462937 0.500000 0.462937 0.367434 0.250000 0.145816 0.072908" << std::endl ;
	of1 << "0.107155 0.214311 0.367434 0.540030 0.680395 0.734867 0.680395 0.540030 0.367434 0.214311 0.107155" << std::endl ;
	of1 << "0.135007 0.270015 0.462937 0.680395 0.857244 0.925875 0.857244 0.680395 0.462937 0.270015 0.135007" << std::endl ;
	of1 << "0.145816 0.291632 0.500000 0.734867 0.925875 1.000000 0.925875 0.734867 0.500000 0.291632 0.145816" << std::endl ;
	of1 << "0.135007 0.270015 0.462937 0.680395 0.857244 0.925875 0.857244 0.680395 0.462937 0.270015 0.135007" << std::endl ;
	of1 << "0.107155 0.214311 0.367434 0.540030 0.680395 0.734867 0.680395 0.540030 0.367434 0.214311 0.107155" << std::endl ;
	of1 << "0.072908 0.145816 0.250000 0.367434 0.462937 0.500000 0.462937 0.367434 0.250000 0.145816 0.072908" << std::endl ;
	of1 << "0.042525 0.085049 0.145816 0.214311 0.270015 0.291632 0.270015 0.214311 0.145816 0.085049 0.042525" << std::endl ;
	of1 << "0.021262 0.042525 0.072908 0.107155 0.135007 0.145816 0.135007 0.107155 0.072908 0.042525 0.021262" << std::endl ;
#else
	/*
	printf("CONV NORM\n") :
	fwhm := 12.0 :
	for pxx from -11 to 11 do
	for pxy from -11 to 11 do
	        printf("%f ",exp(-4.0*log(2.0)*(pxx^2+pxy^2)/fwhm^2)) ;
	end do:
	printf("\n") ;
	end do:
	*/
of1 << "0.009471 0.014191 0.020459 0.028382 0.037885 0.048660 0.060139 0.071518 0.081837 0.090107 0.095465 0.097321 0.095465 0.090107 0.081837 0.071518 0.060139 0.048660 0.037885 0.028382 0.020459 0.014191 0.009471 " << std::endl ;
of1 << "0.014191 0.021262 0.030654 0.042525 0.056764 0.072908 0.090107 0.107155 0.122616 0.135007 0.143035 0.145816 0.143035 0.135007 0.122616 0.107155 0.090107 0.072908 0.056764 0.042525 0.030654 0.021262 0.014191 " << std::endl ;
of1 << "0.020459 0.030654 0.044194 0.061308 0.081837 0.105112 0.129907 0.154487 0.176777 0.194641 0.206215 0.210224 0.206215 0.194641 0.176777 0.154487 0.129907 0.105112 0.081837 0.061308 0.044194 0.030654 0.020459 " << std::endl ;
of1 << "0.028382 0.042525 0.061308 0.085049 0.113527 0.145816 0.180213 0.214311 0.245233 0.270015 0.286071 0.291632 0.286071 0.270015 0.245233 0.214311 0.180213 0.145816 0.113527 0.085049 0.061308 0.042525 0.028382 " << std::endl ;
of1 << "0.037885 0.056764 0.081837 0.113527 0.151541 0.194641 0.240556 0.286071 0.327346 0.360427 0.381859 0.389282 0.381859 0.360427 0.327346 0.286071 0.240556 0.194641 0.151541 0.113527 0.081837 0.056764 0.037885 " << std::endl ;
of1 << "0.048660 0.072908 0.105112 0.145816 0.194641 0.250000 0.308974 0.367434 0.420448 0.462937 0.490465 0.500000 0.490465 0.462937 0.420448 0.367434 0.308974 0.250000 0.194641 0.145816 0.105112 0.072908 0.048660 " << std::endl ;
of1 << "0.060139 0.090107 0.129907 0.180213 0.240556 0.308974 0.381859 0.454109 0.519630 0.572142 0.606163 0.617947 0.606163 0.572142 0.519630 0.454109 0.381859 0.308974 0.240556 0.180213 0.129907 0.090107 0.060139 " << std::endl ;
of1 << "0.071518 0.107155 0.154487 0.214311 0.286071 0.367434 0.454109 0.540030 0.617947 0.680395 0.720853 0.734867 0.720853 0.680395 0.617947 0.540030 0.454109 0.367434 0.286071 0.214311 0.154487 0.107155 0.071518 " << std::endl ;
of1 << "0.081837 0.122616 0.176777 0.245233 0.327346 0.420448 0.519630 0.617947 0.707107 0.778565 0.824861 0.840896 0.824861 0.778565 0.707107 0.617947 0.519630 0.420448 0.327346 0.245233 0.176777 0.122616 0.081837 " << std::endl ;
of1 << "0.090107 0.135007 0.194641 0.270015 0.360427 0.462937 0.572142 0.680395 0.778565 0.857244 0.908218 0.925875 0.908218 0.857244 0.778565 0.680395 0.572142 0.462937 0.360427 0.270015 0.194641 0.135007 0.090107 " << std::endl ;
of1 << "0.095465 0.143035 0.206215 0.286071 0.381859 0.490465 0.606163 0.720853 0.824861 0.908218 0.962224 0.980930 0.962224 0.908218 0.824861 0.720853 0.606163 0.490465 0.381859 0.286071 0.206215 0.143035 0.095465 " << std::endl ;
of1 << "0.097321 0.145816 0.210224 0.291632 0.389282 0.500000 0.617947 0.734867 0.840896 0.925875 0.980930 1.000000 0.980930 0.925875 0.840896 0.734867 0.617947 0.500000 0.389282 0.291632 0.210224 0.145816 0.097321 " << std::endl ;
of1 << "0.095465 0.143035 0.206215 0.286071 0.381859 0.490465 0.606163 0.720853 0.824861 0.908218 0.962224 0.980930 0.962224 0.908218 0.824861 0.720853 0.606163 0.490465 0.381859 0.286071 0.206215 0.143035 0.095465 " << std::endl ;
of1 << "0.090107 0.135007 0.194641 0.270015 0.360427 0.462937 0.572142 0.680395 0.778565 0.857244 0.908218 0.925875 0.908218 0.857244 0.778565 0.680395 0.572142 0.462937 0.360427 0.270015 0.194641 0.135007 0.090107 " << std::endl ;
of1 << "0.081837 0.122616 0.176777 0.245233 0.327346 0.420448 0.519630 0.617947 0.707107 0.778565 0.824861 0.840896 0.824861 0.778565 0.707107 0.617947 0.519630 0.420448 0.327346 0.245233 0.176777 0.122616 0.081837 " << std::endl ;
of1 << "0.071518 0.107155 0.154487 0.214311 0.286071 0.367434 0.454109 0.540030 0.617947 0.680395 0.720853 0.734867 0.720853 0.680395 0.617947 0.540030 0.454109 0.367434 0.286071 0.214311 0.154487 0.107155 0.071518 " << std::endl ;
of1 << "0.060139 0.090107 0.129907 0.180213 0.240556 0.308974 0.381859 0.454109 0.519630 0.572142 0.606163 0.617947 0.606163 0.572142 0.519630 0.454109 0.381859 0.308974 0.240556 0.180213 0.129907 0.090107 0.060139 " << std::endl ;
of1 << "0.048660 0.072908 0.105112 0.145816 0.194641 0.250000 0.308974 0.367434 0.420448 0.462937 0.490465 0.500000 0.490465 0.462937 0.420448 0.367434 0.308974 0.250000 0.194641 0.145816 0.105112 0.072908 0.048660 " << std::endl ;
of1 << "0.037885 0.056764 0.081837 0.113527 0.151541 0.194641 0.240556 0.286071 0.327346 0.360427 0.381859 0.389282 0.381859 0.360427 0.327346 0.286071 0.240556 0.194641 0.151541 0.113527 0.081837 0.056764 0.037885 " << std::endl ;
of1 << "0.028382 0.042525 0.061308 0.085049 0.113527 0.145816 0.180213 0.214311 0.245233 0.270015 0.286071 0.291632 0.286071 0.270015 0.245233 0.214311 0.180213 0.145816 0.113527 0.085049 0.061308 0.042525 0.028382 " << std::endl ;
of1 << "0.020459 0.030654 0.044194 0.061308 0.081837 0.105112 0.129907 0.154487 0.176777 0.194641 0.206215 0.210224 0.206215 0.194641 0.176777 0.154487 0.129907 0.105112 0.081837 0.061308 0.044194 0.030654 0.020459 " << std::endl ;
of1 << "0.014191 0.021262 0.030654 0.042525 0.056764 0.072908 0.090107 0.107155 0.122616 0.135007 0.143035 0.145816 0.143035 0.135007 0.122616 0.107155 0.090107 0.072908 0.056764 0.042525 0.030654 0.021262 0.014191 " << std::endl ;
of1 << "0.009471 0.014191 0.020459 0.028382 0.037885 0.048660 0.060139 0.071518 0.081837 0.090107 0.095465 0.097321 0.095465 0.090107 0.081837 0.071518 0.060139 0.048660 0.037885 0.028382 0.020459 0.014191 0.009471 " << std::endl ;

#endif
	of1.close() ;

	/* temporary file for the output paramter definition.
	*/
	char *skparm = new char[strlen(__FILE__)+13] ;
	sprintf(skparm,"%sXXXXXX.param",__FILE__) ;
	::mkstemps(skparm,6) ;

	/* so the (quieted) output of sex(1) will be a magnitude
	* (useful to sort on), a list of fwhm values (units?) and x and y FITS coordinates
	* of the centroids
	*/
	std::ofstream of2(skparm) ;
	of2 << "MAG_AUTO\n" ;
	of2 << "FWHM_IMAGE\n" ;
	of2 << "X_IMAGE\n" ;
	of2 << "Y_IMAGE\n" ;
	of2.close() ;

	/* temporary file for the image parameters
	*/
	char *sksex = new char[strlen(__FILE__)+11] ;
	sprintf(sksex,"%sXXXXXX.sex",__FILE__) ;
	::mkstemps(sksex,4) ;

	std::ofstream of3(sksex) ;
	// of3 << "# ASSOC_RADIUS 10.0" << std::endl ;
	of3 << "CATALOG_TYPE ASCII" << std::endl ;
	// of3 << "# DETECT_THRESH 18,22" << std::endl ;
	of3 << "DETECT_THRESH " << thresh << std::endl ;
	of3 << "GAIN 2.0" << std::endl ;

	/* pixel scale in arcseconds
	*/
	if ( bin >=1 )
		of3 << "PIXEL_SCALE " << pixscal*bin << std::endl ;

	of3 << "CATALOG_NAME STDOUT" << std::endl ;
	of3 << "PARAMETERS_NAME " << skparm << std::endl ;
	of3 << "SEEING_FWHM 1.2" << std::endl ;
	of3 << "MAG_GAMMA 1.0" << std::endl ;
	of3 << "SATUR_LEVEL 65530" << std::endl ;
	of3 << "BACK_SIZE 100" << std::endl ;
	of3 << "BACK_FILTERSIZE 2" << std::endl ;
	of3 << "THRESH_TYPE RELATIVE" << std::endl ;
	of3 << "DETECT_MINAREA 15" << std::endl ;
	of3 << "FILTER Y" << std::endl ;
	of3 << "FILTER_NAME " << skconv << std::endl ;
	of3 << "MAG_ZEROPOINT 20." << std::endl ;
	if ( batch )
		of3 << "VERBOSE_TYPE QUIET" << std::endl ;
	else
		of3 << "VERBOSE_TYPE FULL" << std::endl ;
		//of3 << "VERBOSE_TYPE NORMAL" << std::endl ;
	of3.close() ;

#if 0
	/* debugging...*/
	std::ostringstream cmd1 ;
	cmd1 << "cat " << sksex ;
	system(cmd1.str().c_str()) ;
#endif
	// cerr << "Not available: sex(1) not found at compile time " << std::endl ;

	/* run sex with output; sort numerically along first column,
	* which places the brightest stars first
	*/
	std::ostringstream cmd ;
	cmd << "sex " << skfits << " -c " << sksex ;
	if ( batch) 
		cmd << " 2>/dev/null | sort -n" ;

	/* open a pipe (which does not receive the parts of the output 
	* with the comments about mean backgrounds etc, but only the lines
	* with the bare 4 parameters (mag, fwhm, x, y)
	*/
	FILE * fp = popen(cmd.str().c_str(),"r") ;

	vector<float> widths ;

	char skout[132] ;
	if ( batch)
	{
		while( fgets(skout,sizeof(skout),fp) != NULL)
		{
			/* read as many paramters as requested above in of2. skout still containts the line feeds.
			* 4 values: magnitude, width, x, y.
			*/
			float tmp[4];
			sscanf(skout,"%f %f %f %f",&tmp[0], &tmp[1], &tmp[2], &tmp[3]) ;
			widths.push_back(tmp[1]) ;

			cout << skout ;
		}
	}
	else
	{
		vector<float> x ;
		vector<float> y ;
		while( fgets(skout,sizeof(skout),fp) != NULL)
		{
			/* read as many paramters as requested above in of2. skout still containts the line feeds.
			* 4 values: magnitude, width, x, y.
			*/
			float tmp[4];
			sscanf(skout,"%f %f %f %f",&tmp[0], &tmp[1], &tmp[2], &tmp[3]) ;
			widths.push_back(tmp[1]) ;
			x.push_back(tmp[2]) ;
			y.push_back(tmp[3]) ;
		}
		ds9(skfits,x,y,widths) ;
	}
	pclose(fp) ;

	/* delete parameter file and the associated name
	* Note that the following cleanup should not be done while the pipe is open!
	*/
	remove(sksex) ;
	delete[] sksex ;

	/* delete parameter file and the associated name
	*/
	remove(skparm) ;
	delete[] skparm ;

	/* delete convolution file and the associated name
	*/
	remove(skconv) ;
	delete[] skconv ;

	return widths ;

} /* Fits2fwhm::centroids */


/*!**************************************** 
* @brief Show the picture with ds9(1).
* @param[in] skfits The FITS file to be shown.
* @param[in] x FITS x coordinates
* @param[in] y FITS y coordinates
* @param[in] w width parameters in some sex(1) units
* @param[in] N number of elements in x, y and w.
* @author Richard J. Mathar
* @since 2020-03-02
*/
void Fits2fwhm::ds9(const char * skfits, const vector<float> &x, const vector<float> &y, const vector<float> & w) const
{
	/* produce a DS9 regions file with pixel coordinates in some suitable
	* temporary ASCII file
	*/
	char *skreg = new char[strlen(__FILE__)+11] ;
	sprintf(skreg,"%sXXXXXX.reg",__FILE__) ;
	::mkstemps(skreg,4) ;

	std::ofstream ofi(skreg) ;
	if ( ! ofi)
	{
		std::cerr<< "Cannot open " << skreg << std::endl ;
		remove(skreg) ;
		delete [] skreg ;
	}
	else
	{
		/* loop thru all stars (this includes all stars in the catalogue, whether
		* inside the GWS ring, inside the HWS ring or entirely outside the FOV...)
		* Convert coordinates from radians to degress; attempt to use 20 pix for the marker,
		* and use the 'type' indicator (first field of the catalogu) as the name of the star.
		* Tests with sim-2016b show that the RA/DEC coordinates of the TCS are not (!) precessed
		* to the current date, but only advanced by proper motion. To match this we do not (!)
		* precess our stars either, but apply only the proper motions.
		*/
		for(int s=0 ; s < (int)x.size() ; s++)
		{
			ofi << "image;box " << x[s] << " " << y[s]
				<< " 10i 10i # color=yellow width=3 dash=1"
				<< " text={" << w[s] << "}"
				<< std::endl ;
		}

		ofi.close() ;
        }

#if 0
	// debugging the regions file...
	std::ostringstream cmd1 ;
	cmd1 << "cat " << skreg ;
	system(cmd1.str().c_str()) ;
#endif

	/* run ds9 on the FITS file
	*/
	std::ostringstream cmd ;
	cmd << "ds9 -view vertical" ;

	cmd << " -geometry 640x540" ;

	cmd << " " << skfits ;

	cmd << " -zoom 0.7"
        	<< " -scale mode 98"
		//<< " -cmap BB "
		// << " -title ' mockup SE " << sdx.toStringCfg() << " HWS"
		//<< " " << now.toString(0,false) << " loc. time'"
		// << " -grid yes"
		<< " -grid no"
		//<< " -crosshair match"
		<< " -colorbar no"
		<< " -view filename no"
#if 0
		/* crosshair prefered probably by experienced people, but not for first-time users....
		*/
		<< " -mode crosshair"
#endif
		//<< " -view wcsa yes"
		//<< " -view wcs no"
		<< " -view physical no"
		<< " -regions load " << skreg ;

	// cmd << " -rotate " << seRotang ;
	// cmd << " -orient y" ;
	/* we don't put this ds9 command into the background with & because that would
	* delete the FITS file before ds9 is done with it...
	*/
	system(cmd.str().c_str()) ;

	remove(skreg) ;
	delete [] skreg ;

} /* ds9() */

/*!**************************************** 
* @brief Extract a subregion of a FITS file
* @param[in] infile The file name of the original FITS file.
*   This supports the cfitsio convention of specifying suregions in brackets.
* @param[in] outfile The name of the file to be (over)written.
* @param[in] copyall If nonzero, copy all FITS header keywords.
* @return nonzero cfitsio status if something went wrong
* @author Richard J. Mathar
* @since 2021-06-02
*/
int Fits2fwhm::ftcopy(const char * infile, const char *outfile, int copyall)
{
	/*  copy the input HDU or file to the output file */
	fitsfile *infptr = 0, *outfptr = 0;
	int status = 0;

    	/* Open the input file, and move to first 'interesting' HDU */
	if (fits_open_data(&infptr, infile, READONLY, &status)) 
	{
		fits_close_file(infptr,  &status);
		return status ;
	}

	unlink(outfile);  /* delete existing file if clobber=YES */

    	if (fits_create_file(&outfptr, outfile, &status)) 
	{
		fits_close_file(outfptr, &status);
		fits_close_file(infptr,  &status);
		return status ;
	}

	if (copyall) 
	{   
		/* copying all HDUs? */

		if (fits_copy_file(infptr, outfptr, 1, 1, 1, &status)) 
		{
			fits_close_file(outfptr, &status);
			fits_close_file(infptr,  &status);
			return status ;
		}

		/* move back to 1st HDU, before writing HISTORY keywords */
		if (fits_movabs_hdu(outfptr, 1, NULL, &status)) 
		{
			fits_close_file(outfptr, &status);
			fits_close_file(infptr,  &status);
			return status ;
		}

	} 
	else 
	{   
		/* just copy the current HDU */

		if (fits_copy_hdu(infptr, outfptr, 0, &status)) 
		{
			fits_close_file(outfptr, &status);
			fits_close_file(infptr,  &status);
			return status ;
		}
	}

	/* write optional history keywords 
	* status = HDpar_stamp(outfptr, 0, &status);
	*/

	fits_close_file(outfptr, &status);
	fits_close_file(infptr,  &status);

	return status ;
} /* ftcopy */
