/*
 $Header: https://svn.mpia.de/gulli/geirs/src/trunk/extern/fits2fwhm/fitSqr.cxx 798 2021-06-15 13:53:57Z mathar $
*/

#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <iostream>
#include <fstream>
#include <sstream>
#include <limits>
#include <unistd.h>
#include <errno.h>

#include "FitSqr.h"

using namespace std ;

/** Call gnuplot(1) to show the scattered input data and fitted parabola
* @param x The array of x values.
* @param y The array of y values.
* @param N The number of values in x[] and also in y[].
* @param lincoef The three floating point coefficients of the 2nd order polynomial
*   The polynomial is lincoef[0]+lincoef[1]*x+lincoef[2]*x^2.
*/
void callGnuplot(const double *x, const double *y , int N, const double * lincoef)
{
	/* generate a temporary file with
	*  plot "fname", linceof[0]+linceof[1]+linceof[2]*x**2 notitle
	* and call that as executable.
	* Optionally (better) use the inline data blocks...
	*/
	char tmpf[] = "tmpXXXXXX" ;
	mkstemp(tmpf) ;
	ofstream gpl(tmpf) ;

	gpl << "$xy << EOD" << endl ;
	for (int i=0; i < N ; i++)
		gpl << x[i] << " " << y[i] <<endl ;
	gpl << "EOD" << endl ;
	const double xapex = -0.5*lincoef[1]/lincoef[2] ;
	gpl << "set xlabel \"apex at " << xapex << "\"" << endl ;

	gpl << "plot $xy notitle, " << lincoef[0] << std::showpos << lincoef[1] << "*x" << std::showpos << lincoef[2] << "*x**2" 
			<< " title \"press q to close\"" << endl ;
	gpl << "pause -1" << endl ;

	ostringstream cmd ;
	cmd << "gnuplot -c " << tmpf ;
	system(cmd.str().c_str()) ;

	unlink(tmpf) ;
}

/** Main demonstrator function.
* The program is called with one argument, which is an existing
* ASCII file. This file contains either lines starting with #, and these
* lines are ignored. Other lines contain each two floating point numbers,
* separated by white space, which define one pair of x and y.
*
* The function prints the 3 coefficients of the quadratic fit, the mean
* error of this fit and then the 3 coefficients of the fit with the rational
* polynomial and its mean error.
* @since 2012-12-04
*/
int main(int argc, char *argv[])
{
	// extern int errno ;

	/* number of input pairs
	*/
	int N=0 ;

	/* quiet or interactive?
	*/
	bool quiet = false ;

	/* the three coefficients of the fit
	*/
	double lincoef[3] ;

	/* variances of the three coefficients of the fit
	*/
	double lincoefVar[3] ;

	/* vector of x and y inputs
	*/
	double *x = (double *)malloc(0) ;
	double *y = (double *)malloc(0) ;

	/* the range of the x values of the inputs
	* Initialized with the "opposite" values
	* such that any upcoming change of the minimum will make it smaller
	* and any change of the maximum will make it larger.
	*/
	double xmimax[2] ;
	xmimax[0] = numeric_limits<double>::max() ;
	xmimax[1] = numeric_limits<double>::min() ;

	/* 0-based column indices for the x and y values
	*/
	int xycol[2] = {0,1} ;

	/* command line option character
        */
        char oc;
        while (  (oc=getopt(argc,argv,"qx:y:")) != -1 )
        {
                switch(oc)
                {
                case 'q' :
			quiet = true ;
                        break ;
                case 'x' :
			xycol[0] = atoi(optarg)-1 ;
                        break ;
                case 'y' :
			xycol[1] = atoi(optarg)-1 ;
                        break ;
                case '?' :
                        cerr << "Warning: Unsupported command line option " << optopt << " ... ignored" << endl ;
                        break ;
                }
        }

	if ( optind >= argc)
	{
		cerr << "Usage: " << argv[0] << " [-q] [-x colxvalues] [-y colyvalues] inputfile.asc" << endl ;
		return 1 ;
	}

	if ( xycol[0] < 0 )
	{
		cerr << "Error: invalid column index " << 1+xycol[0] << endl ;
		return 1 ;
	}

	if ( xycol[1] < 0 )
	{
		cerr << "Error: invalid column index " << 1+xycol[1] << endl ;
		return 1 ;
	}

	if ( xycol[0] == xycol[1])
	{
		cerr << "Error: same column index " << 1+xycol[0] << " for x and y values" << endl ;
		return 1 ;
	}

	/* open file and read all lines that are not starting with
	* the sharp. Slow single scan, certainly not optimized for efficiency.
	*/
	ifstream fin(argv[optind]) ;
	if ( ! fin )
	{
		cerr << "Cannot open " << argv[optind] << " " << strerror(errno) << endl ;
		return 1 ;
	}

	while( !fin.eof() )
	{
		/* the current line of the input file */
		string inl ;
		getline(fin,inl) ;
		if( inl[0] != '#' && inl.length() >=3 )
		{
			/* skip comments and empty lines (minimum length is 3:
			* the x and y values have one digit and are separated by a single blank..)
			*/
			double xy[2] = { numeric_limits<double>::quiet_NaN(), numeric_limits<double>::quiet_NaN() } ;
			/* convert the line to a string stream
			*/
			istringstream inpstr(inl) ;

			/* parse string left-right; skip fields/columns not in the xycol spec
			*/
			for(int f=0 ; f <= max(xycol[0],xycol[1]) ; f++)
			{
				string field ;
				inpstr >> field ;
				errno =0 ;
				if ( f == xycol[0] && field.length() )
				{
					xy[0] = atof(field.c_str()) ;
					if ( errno )
						xy[0] = numeric_limits<double>::quiet_NaN() ;
				}
				else if ( f == xycol[1] && field.length() )
				{
					xy[1] = atof(field.c_str()) ;
					if ( errno )
						xy[1] = numeric_limits<double>::quiet_NaN() ;
				}
			}

			/* got a valid input line?
			*/
			if ( ! isnan(xy[0]) && ! isnan(xy[1]) )
			{
				N++ ;
				x= (double*)realloc(x,N*sizeof(double)) ;
				y= (double*)realloc(y,N*sizeof(double)) ;
				x[N-1] = xy[0] ;
				y[N-1] = xy[1] ;
				xmimax[0] = std::min(xmimax[0],xy[0]) ;
				xmimax[1] = std::max(xmimax[1],xy[0]) ;
			}
			else
			{
				cerr << "skipping invalid input " << inl << endl ;
			}
		}
	}
	fin.close() ;

	if ( N < 3 )
	{
		cerr << "Insufficient " << N << " data pairs in " << argv[optind] << endl ;
		free(x) ;
		free(y) ;
		return 1 ;
	}

	/* do the quadratic fit */
	FitSqr::fitsqr(x,y,N,lincoef, lincoefVar) ;
	/* now we have the fit y=lincoef[0]+lincoef[1]*x+lincoef[2]*x^2, and
	* the minimum is at -lincoef[1]/2/lincoef[2].
	*/
	const double xapex = -0.5*lincoef[1]/lincoef[2] ;
	if ( quiet )
	{
		cout << xapex << endl ;
	}
	else
	{
		cout << xapex <<" x apex " << endl ;
		cout << lincoef[0] << " " << lincoef[1] << " " << lincoef[2] << " poly coefficients" << endl ;
		callGnuplot(x,y,N,lincoef) ;

		/* The standard application here is a focal series
		*  and we expect y to be fwhm values and whish to reduce these to
		* a point of minimum y, not maximum y.
		*/
		if ( lincoef[2] < 0.0 )
			cerr << "Warning: parabola with maximum, not minimum" << endl ;
		if ( xapex < xmimax[0] || xapex > xmimax[1] )
			cerr << "Warning: apex " << xapex << " outside scan interval " 
				<< xmimax[0] << ".." << xmimax[1] << endl ;
	}


	/* free resources bundled above */
	free(x) ;
	free(y) ;
	return 0 ;
} /* main */
